/**
 * XPathTester v1.4 (Saxon)
 * Copyright (c) 2001 by FiveSight Technologies, Inc. with contributions from
 * (in alphabetical order) Paul Brown, Justin Guinney, and Maciej Szefler.
 *
 * Visit FiveSight on the web at "http://www.fivesight.com/".
 *
 * This software is covered by the FiveSight Public Software License, as
 * follows:
 *
 * ============================================================================
 * The FiveSight Public Software License
 *
 * Copyright (c) 2001 FiveSight Technologies, Inc.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by
 *       FiveSight Technologies, Inc. (http://www.fivesight.com/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "FiveSight" and "FiveSight Technologies" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact info@fivesight.com.
 *
 * 5. Products derived from this software may not be called "FiveSight",
 *    nor may "FiveSight" appear in their name, without prior written
 *    permission of FiveSight Technologies.
 *
 * ============================================================================
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL FIVESIGHT TECHNOLOGIES, INC. BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
 * OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
 * PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF
 * THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * ============================================================================
 */

package com.fs.xpathtester;

/**
 * Xalan J2 XPath support
 */
//import org.apache.xpath.*;
//import org.apache.xml.utils.*;
//import org.apache.xpath.objects.*;

/**
 * Saxon 6 XPath support
 */
import com.icl.saxon.*;
import com.icl.saxon.expr.*;
import com.icl.saxon.om.*;
import com.icl.saxon.tree.*;

import javax.xml.transform.*;
import javax.xml.transform.sax.*;


import org.xml.sax.*;
import org.w3c.dom.*;
//import org.w3c.dom.traversal.*;
import javax.xml.parsers.*;

import javax.swing.*;
import javax.swing.event.*;
import javax.swing.tree.*;
import javax.swing.text.*;
import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;
import java.net.*;

/**
 * This JFrame displays an XML document in a JTree with a JTextField for entering
 * an XPath query.
 */
public class XPathTester extends JFrame {

  /** Match highlight color for the JTree nodes. */
  private static final java.awt.Color XPATH_MATCHED_COLOR = Color.yellow;
  /** Context node text color. */
  private static final java.awt.Color CONTEXT_COLOR = Color.blue;
  /** Normal background color for the JTree nodes. */
  private static final java.awt.Color NORMAL_COLOR = Color.white;

  /** Error highlight color for the XPath text entry field. */
  private static final Color ERROR_COLOR = Color.pink;
  /** Normal background color for the XPath text entry field. */
  private static final Color OK_COLOR = Color.white;


  org.w3c.dom.Document _document = null;

  /*
   * Xalan J2 XPath support.
   */
//  private Node context = null;
//  private XPathAPI xp;
//  private PrefixResolverDefault _pr;

  /**
   * Saxon 6 XPath support.
   */
  private Controller _controller;
  private Context _controllerContext; // evaluation context
  private StaticContext _staticContext; // variable bindings
  private VariableReference[] _vars;
  private DocumentInfo _di;

  DefaultMutableTreeNode _contextTreeNode;

  Vector treeNodes = new Vector();

  /**
   * JTextField extension used to capture XPath input.
   */
  XPathField xpathTextField;

  /**
   * JFileChooser for XML files.
   */
  private javax.swing.filechooser.FileFilter xmlFiles = new javax.swing.filechooser.FileFilter() {
    public boolean accept(File f) {
      return f != null && (f.isDirectory() || f.getName().toUpperCase().endsWith(".XML"));
    }

    public String getDescription() {
      return "XML files (.XML, .xml, etc.)";
    }
  };
  private JFileChooser fileChooser;

  /**
   * JAXP XML parser support.
   */
//  private DocumentBuilderFactory dbf;
//  private DocumentBuilder db;
  SAXParserFactory _spf;
  SAXParser _sp;

  /**
   * Panes
   */
  private JScrollPane treeScrollPane = null;
  private JPanel controlPane = new JPanel(new BorderLayout());
  private JPanel resultPane = new JPanel(new BorderLayout());
  private JPanel topPane = new JPanel(new BorderLayout());;
  private JPanel mainContentPane = new JPanel(new BorderLayout());
  private Box buttonBox = Box.createHorizontalBox();

  /*
   * Menus
   */
  JMenuBar menuBar = new JMenuBar();
  JMenu fileMenu = new JMenu("File");
  JMenuItem help = new JMenu("Help");
  JMenuItem about = new JMenuItem("About");
  JMenuItem xmlDocMenuItem = new JMenuItem("Xml file...");
  JMenuItem exitMenuItem = new JMenuItem("Exit");

  /*
   * JCheckbox and JLabel stuff.
   */
  private static final String nonRT = "Hit enter to execute query:";
  private static final String rt = "Enter query interactively:";
  private JLabel instructionLabel = new JLabel(nonRT);
  private JPanel instructionBox = new JPanel(new BorderLayout());
  private JLabel resultCount = new JLabel("RESULT: No query run yet.");
  private JCheckBox realTimeCBox = new JCheckBox("Real-time query",false);

  /*
   * Catch-all ActionListener
   */
  private XPathTesterActionListener xpathTestActionListener =
    new XPathTesterActionListener();

  /**
   * DOMTree to display the XML document.
   */
  private DOMTree _domTree;

  /**
   * Create a new, empty XPathTester JFrame.
   */
  public XPathTester() {
    super("XPath Tester");
    //xp = new XPathAPI();
    _spf = SAXParserFactory.newInstance();
    // Enable namespace support.
    _spf.setNamespaceAware(true);

    try {
      _sp = _spf.newSAXParser();
    } catch (Exception e) {
      JOptionPane.showMessageDialog(this,
        "Unable to obtain XML parser: " + e.getMessage(),
        "Fatal Error",JOptionPane.ERROR_MESSAGE);
      System.exit(-1);
    }
    initGui();
  }

  /* Load an XML document */
  private org.w3c.dom.Document loadXMLdocument() {
    int ret = fileChooser.showOpenDialog(this);
    File f = fileChooser.getSelectedFile();
    if (f == null || f.isDirectory()) return null;
    try {
      // standalone use of Saxon's facilities requires some gymnastics.
      SAXSource s = new SAXSource(new InputSource(new FileInputStream(f)));
      _controller = new Controller();
      _controller.setTreeModel(Builder.STANDARD_TREE);
      Builder b = _controller.makeBuilder();
      b.build(s);
      DocumentInfo di = b.getCurrentDocument();
      _controllerContext = _controller.makeContext(di);
      _staticContext = new StandaloneContext(_controller.getNamePool());
      this.setTitle("XPath Tester - " + f.getAbsolutePath());
      _domTree.clearSelection();
      resultCount.setText("No query executed.");
      _domTree.invalidate();
      _contextTreeNode = null;
      _di = di;
      return (org.w3c.dom.Document) di;
    } catch (Exception e) {
      JOptionPane.showMessageDialog(this,"Unable to load file: " + e.getMessage(),
        "Error Loading XML file", JOptionPane.ERROR_MESSAGE);
      return null;
    }
  }

  /* Create the GUI. */
  private void initGui() {
    /* JMenu setup */
    setJMenuBar(menuBar);
    menuBar.add(fileMenu);
    menuBar.add(help);
    help.add(about);
    fileMenu.add(xmlDocMenuItem);
    fileMenu.addSeparator();
    fileMenu.add(exitMenuItem);
    /* "Real-time" checkbox */
    realTimeCBox.addChangeListener(new ChangeListener() {
      public void stateChanged(ChangeEvent e) {
        instructionLabel.setText(realTimeCBox.isSelected()?rt:nonRT);
        xpathTextField.setRealTime(realTimeCBox.isSelected());
        if (realTimeCBox.isSelected()) { analyze(); }
      }
    });
    /* XPath input field */
    xpathTextField = new XPathField(40,this);
    xpathTextField.addActionListener(xpathTestActionListener);
    xpathTextField.addKeyListener(new KeyAdapter(){
      /* Grab the [ENTER] key and [TAB] key */
      public void keyPressed(KeyEvent evt){
        boolean eot = evt.getKeyCode() == KeyEvent.VK_TAB ||
          evt.getKeyCode() == KeyEvent.VK_ENTER;
        if (eot) {
          analyze();
          if (evt.getKeyChar() == KeyEvent.VK_ENTER)
            analyze();
            evt.consume();
        }
      }
    });
    /* Hook up the ActionListener */
    xmlDocMenuItem.addActionListener(xpathTestActionListener);
    exitMenuItem.addActionListener(xpathTestActionListener);
    about.addActionListener(xpathTestActionListener);
    /* Setup for the DOMTree. */
    _domTree = new DOMTree(true,true,true);
    treeScrollPane = new JScrollPane(_domTree);
    _domTree.setCellRenderer(new DomTreeCellRenderer());
    treeScrollPane.setPreferredSize(new Dimension(400,300));
    /* Populate the JFrame. */
    instructionBox.add(instructionLabel,BorderLayout.WEST);
    instructionBox.add(realTimeCBox,BorderLayout.EAST);
    topPane.add(instructionBox,BorderLayout.NORTH);
    topPane.add(xpathTextField,BorderLayout.CENTER);
    resultPane.add(resultCount,BorderLayout.WEST);
    topPane.add(resultPane,BorderLayout.SOUTH);
    mainContentPane.add(topPane, BorderLayout.NORTH);
    mainContentPane.add(treeScrollPane, BorderLayout.CENTER);
    /* Handle the selection of a base context via a click on the tree. */
    _domTree.addTreeSelectionListener(new TreeSelectionListener() {

      NodeInfo _last = null;

      public void valueChanged(TreeSelectionEvent e) {
        DefaultMutableTreeNode treeNode = (DefaultMutableTreeNode)
          _domTree.getLastSelectedPathComponent();
        if (treeNode == null) return;
        NodeInfo ni = (NodeInfo) _domTree.getNode(treeNode);
        if (treeNode.toString().startsWith("</")) {
          _domTree.setSelectionPath(
            _domTree.getSelectionPath().getParentPath().pathByAddingChild(
              treeNode.getPreviousSibling()
            )
          );
          return;
        }
        if (ni == _last) {
          _controllerContext = _controller.makeContext(ni.getDocumentRoot());
          _last = null;
          _contextTreeNode = null;
          _domTree.clearSelection();
          analyze();
          _domTree.repaint();
          return;
        } else {
          _controllerContext = _controller.makeContext(ni);
        }
        _contextTreeNode = treeNode;
        analyze();
        _last = ni;
        _domTree.clearSelection();
        _domTree.repaint();
      }
    });
    DefaultTreeSelectionModel dftsm = new DefaultTreeSelectionModel();
    dftsm.setSelectionMode(DefaultTreeSelectionModel.DISCONTIGUOUS_TREE_SELECTION);
    _domTree.setSelectionModel(dftsm);
    /* Finish setup. */
    JPanel panel = new JPanel(new GridLayout(1,1));
    panel.setBorder(BorderFactory.createEmptyBorder(5,5,5,5));
    panel.add(mainContentPane);
    this.getContentPane().add(panel);
    fileChooser = new JFileChooser();
    fileChooser.setFileFilter(xmlFiles);
    _document = loadXMLdocument();
    _domTree.setDocument(_document);
  }

  /**
   * Apply an XPath to the tree.
   */
  public void analyze(){
    try {
      // remove any marks from a previous query
      treeNodes.clear();
      String xpath_str = xpathTextField.getText();
      if(xpath_str == null || xpath_str.equals("") || _document == null) {
        resultCount.setText("No expression or no document loaded.");
        return;
      }
      Expression e = Expression.make(xpath_str,_staticContext);
      Value v = e.evaluate(_controllerContext);
      // Determine the display according to the result type.
      switch (v.getDataType()) {
        case Value.NODESET:
          int len = showXpathResults(v.enumerate(_controllerContext,false));
          resultCount.setText("Query returned: " + len + " match" + ((len==1)?"":"es"));
          _domTree.repaint();
          break;
        case Value.NUMBER:
          double d = v.asNumber();
          String result =
            Math.round(d) == d ? String.valueOf((int) Math.round(d)):String.valueOf(d);
          resultCount.setText("Query returned a number: " + result);
          break;
        case Value.STRING:
          resultCount.setText("Query returned a string: " + v.asString());
          break;
        case Value.BOOLEAN:
          resultCount.setText("Query returned a boolean: " + v.asBoolean());
          break;
      }
      xpathTextField.setBackground(OK_COLOR);
    } catch (Exception ex) {
      // This is a little dicey -- parse the exception message to find something
      // to display.
      String s = ex.getMessage();
      if (s == null || s.length() == 0) {
        resultCount.setText("Unknown error. (" + ex.getClass().getName() + ")");
      } else {
        int l = s.indexOf("\n");
        if (l == -1) {
          resultCount.setText("ERROR: " + ex.getMessage());
        } else {
          if (l>s.length() - 2) {
            resultCount.setText("ERROR: " + ex.getMessage().substring(0,l-1));
          } else {
            StringBuffer sb = new StringBuffer(ex.getMessage().substring(l+1));
            for (int i=0;i < sb.length(); ++i)
              if (Character.isWhitespace(sb.charAt(i)))
                sb.setCharAt(i,' ');
            resultCount.setText("ERROR: " + sb.toString());
          }
        }
      }
      // Notify the user via a color change.
      xpathTextField.setBackground(ERROR_COLOR);
    }
  }

  /**
   * Highlight a NodeSet in the tree based on a supplied NodeIterator; be sure
   * to expand the tree as necessary to expose the target nodes.
   *
   * @return the number of nodes in the iterator.
   */
  private int showXpathResults(NodeEnumeration ne) {
    DOMTree.Model model = (DOMTree.Model)_domTree.getModel();
    int cnt = 0;
    Node n;
    while (ne.hasMoreElements()) {
      try {
        MutableTreeNode treeNode = model.getTreeNode(n = (Node) ne.nextElement());
        if(treeNode != null) {
          _domTree.expandPath(
            (new TreePath(model.getPathToRoot(treeNode))).getParentPath());
          treeNodes.addElement(treeNode);
        }
        ++cnt;
      } catch (XPathException xpe) { }
    }
    if(cnt != 0) _domTree.repaint();
    return cnt;
  }

  /**
   * Fancy JTextField subclass to provide incremental execution of the XPath query
   * and highlighting of errors.
   */
  protected class XPathField extends JTextField {

    private XPathTester parent = null;
    private boolean realTime = false;

    // Setup is vanilla.
    public XPathField(int cols, XPathTester owner) {
      super(cols);
      parent = owner;
    }

    public void setRealTime(boolean flag) {
      realTime = flag;
    }

    // Call-back to containing JFrame
    public void setOwner(XPathTester owner) {
      parent = owner;
    }

    protected javax.swing.text.Document createDefaultModel() {
      return new XPathDocument();
    }

    /**
     * Subclass on javax.swing.text.PlainDocument to capture insertions.
     */
    protected class XPathDocument extends javax.swing.text.PlainDocument {
      // send insertions back up to the containing XPathTester
      public void insertString(int offset, String str,
        javax.swing.text.AttributeSet a)
        throws javax.swing.text.BadLocationException
      {
        if (str == null) return;
        super.insertString(offset,str,a);
        if (realTime && parent != null)  parent.analyze();
      }

      // send removals back up to the containing XPathTester
      public void remove(int offset, int len)
        throws javax.swing.text.BadLocationException
      {
        super.remove(offset,len);
        if (realTime && parent != null) parent.analyze();
      }
    }
  }

  protected class XPathTesterActionListener implements ActionListener
  {
    public void actionPerformed(ActionEvent evt)
    {
      Component comp = (Component)evt.getSource();
      if(comp == xmlDocMenuItem) {
        _document = loadXMLdocument();
        _domTree.setDocument(_document);
      } else if(comp == xpathTextField) { analyze(); }
      else if(comp == about) { showAbout(); }
      else if(comp == exitMenuItem) { System.exit(0); }
    }
  }

  /* Set up for fancy "About" box. */
  private static final SimpleAttributeSet[] sas = new SimpleAttributeSet[4];
  static {
    sas[0] = new SimpleAttributeSet();
    StyleConstants.setFontFamily(sas[0],"Sans");
    StyleConstants.setFontSize(sas[0],16);
    StyleConstants.setBold(sas[0],true);
    sas[1] = new SimpleAttributeSet(sas[0]);
    StyleConstants.setFontSize(sas[1],12);
    StyleConstants.setBold(sas[1],false);
    sas[2] = new SimpleAttributeSet(sas[1]);
    StyleConstants.setFontFamily(sas[2],"Serif");
    StyleConstants.setFontSize(sas[2],9);
    StyleConstants.setForeground(sas[2],Color.blue);
    sas[3] = new SimpleAttributeSet(sas[2]);
    StyleConstants.setForeground(sas[3],Color.red);
    StyleConstants.setBackground(sas[3],Color.white);
  }

  private static final DefaultStyledDocument aboutDoc = new DefaultStyledDocument();
  static {
    appendTo(aboutDoc,"XPathTester",0);
    appendTo(aboutDoc,"Version 1.4 for Saxon (3 December 2001)",1);
    appendTo(aboutDoc,"Copyright (c) 2001 by FiveSight Technologies, Inc.",1);
    appendTo(aboutDoc,"URL:   http://www.fivesight.com",2);
    appendTo(aboutDoc,"Email: devtips@fivesight.com\n\n",2);
    appendTo(aboutDoc,"XPathTester is free software covered by the FiveSight Public Software License; you may copy, modify, and/or redistribute them under certain conditions.  See the \"FiveSight PSL\" tab for details.\n",1);
    appendTo(aboutDoc,"XPath Tester comes with ABSOLUTELY NO WARRANTY.  If this is somehow unclear, see the \"FiveSight FPSL\" tab for details.\n",1);
    appendTo(aboutDoc,"This version of the XPathTester includes version 6.5 of the Saxon XSLT processor.  Saxon is free software covered by the Mozilla Public License; see the \"Mozilla MPL\" tab for details.",1);
  }

  private DefaultStyledDocument FPSLDoc = null;
  private DefaultStyledDocument MPLDoc = null;
  private JDialog aboutDialog = null;

  private static void appendTo(StyledDocument d, String s, int i) {
    try {
      d.insertString(d.getLength(),s + "\n",sas[i]);
    } catch (BadLocationException ble) {}
  }

  private DefaultStyledDocument errorOnLoad(Exception e) {
    DefaultStyledDocument d = new DefaultStyledDocument();
    appendTo(d,"Unexpected exception loading resource:",0);
    appendTo(d,e.getMessage(),3);
    appendTo(d,"",3);
    StringWriter sw = new StringWriter();
    e.printStackTrace(new PrintWriter(sw));
    appendTo(d,sw.getBuffer().toString(),3);
    return d;
  }

  private DefaultStyledDocument createTextPane(URL u, String title) {
    try {
      InputStream is = u.openStream();
      int c;
      StringBuffer sb = new StringBuffer();
      while ((c=is.read()) != -1) sb.append((char)c);
      is.close();
      DefaultStyledDocument d = new DefaultStyledDocument();
      appendTo(d,title,0);
      appendTo(d,sb.toString(),2);
      return d;
    } catch (Exception e) {
      return errorOnLoad(e);
    }
  }

  /**
   * Fancy About dialog.
   */
  public void showAbout() {
    if (aboutDialog == null) {
      URLClassLoader cl = (URLClassLoader)getClass().getClassLoader();
      if (FPSLDoc == null) {
        URL fpsl = cl.findResource("fpsl.txt");
        FPSLDoc = createTextPane(fpsl,"FiveSight Public Software License");
      }
      if (MPLDoc == null) {
        URL mpl = cl.findResource("mpl.txt");
        MPLDoc = createTextPane(mpl,"Mozilla Public License, v. 1.1");
      }
      JTextPane jta = new JTextPane(aboutDoc);
      jta.setEditable(false);
      JScrollPane a = new JScrollPane(jta);
      JTextPane jtb = new JTextPane(FPSLDoc);
      jtb.setEditable(false);
      JScrollPane b = new JScrollPane(jtb);
      JTextPane jtc = new JTextPane(MPLDoc);
      jtc.setEditable(false);
      JScrollPane c = new JScrollPane(jtc);

      JTabbedPane jtp = new JTabbedPane(JTabbedPane.BOTTOM);
      jtp.addTab("About",a);
      jtp.addTab("FiveSight PSL",b);
      jtp.addTab("Mozilla MPL",c);
      aboutDialog = new JDialog(this,"About XPath Tester",true);
      aboutDialog.getContentPane().add(jtp);
      aboutDialog.setSize(new Dimension(400,400));
    }
    aboutDialog.setLocationRelativeTo(this);
    aboutDialog.setVisible(true);
  }

  /**
   * DefaultTreeCellRenderer extension to highlight matched nodes in the DOMTree.
   */
  class DomTreeCellRenderer extends DefaultTreeCellRenderer {
    public Component getTreeCellRendererComponent(JTree tree, Object value,
      boolean sel, boolean expanded, boolean leaf, int row, boolean hasFocus)
    {
      super.getTreeCellRendererComponent(tree, value, sel, expanded, leaf,
        row, hasFocus);
      DefaultMutableTreeNode treeNode = (DefaultMutableTreeNode)value;
      setIcon(null);
      if(treeNodes.contains(treeNode)) {
        setBackgroundNonSelectionColor(XPATH_MATCHED_COLOR);
      } else {
        setBackgroundNonSelectionColor(NORMAL_COLOR);
      }
      if(treeNode == _contextTreeNode) {
        setForeground(CONTEXT_COLOR);
        setBorder(BorderFactory.createLineBorder(Color.blue));
        setMinimumSize(getPreferredSize());
      } else {
        setBorder(BorderFactory.createEmptyBorder());
        setForeground(Color.black);
      }
      return this;
    }
  }

  /**
   * Standalone invocation method.
   */
  public static void main(String[] argv) {
    JFrame frame = new XPathTester();
    frame.setSize(400,500);
    frame.addWindowListener(new WindowAdapter() {
      public void windowClosing(WindowEvent we) {
        System.exit(-1);
      }
    });
    frame.setVisible(true);
  }

}
